@@ -22,6 +22,19 @@ import pdb
 
 from bigjob import SAGA_BLISS 
 from bigjob.state import Running, New, Failed, Done, Unknown
+
+# Optional Job Plugins
+try:
+    from .job_plugin.gcessh import Service as GCEService
+except:
+    pass 
+
+try:
+    from .job_plugin.ec2ssh import Service as EC2Service
+except:
+    pass 
+
+
 # import other BigJob packages
 # import API
 import api.base
@@ -175,7 +188,6 @@ class bigjob(api.base.bigjob):
             
     def start_pilot_job(self, 
                  lrms_url, 
-                 bigjob_agent_executable=None,
                  number_nodes=1,
                  queue=None,
                  project=None,
@@ -184,7 +196,8 @@ class bigjob(api.base.bigjob):
                  walltime=None,
                  processes_per_node=1,
                  filetransfers=None,
-                 external_queue=""):
+                 external_queue="",
+                 pilot_compute_description=None):
         """ Start a batch job (using SAGA Job API) at resource manager. Currently, the following resource manager are supported:
             fork://localhost/ (Default Job Adaptor
             gram://qb1.loni.org/jobmanager-pbs (Globus Adaptor)
@@ -201,6 +214,7 @@ class bigjob(api.base.bigjob):
         lrms_saga_url = SAGAUrl(lrms_url)
         self.url = lrms_saga_url
         self.pilot_url = self.app_url + ":" + lrms_saga_url.host
+        self.number_nodes=int(number_nodes)*int(processes_per_node)  
         
         # Store references to BJ in global dict
         _pilot_url_dict[self.pilot_url]=self
@@ -210,10 +224,19 @@ class bigjob(api.base.bigjob):
         self.coordination.set_pilot_state(self.pilot_url, str(Unknown), False)
         self.coordination.set_pilot_description(self.pilot_url, filetransfers)    
         logger.debug("set pilot state to: " + str(Unknown))
+        
         ##############################################################################
+        # Create Job Service (Default: SAGA Job Service, alternative Job Services supported)
+        self.js =None
+        if lrms_saga_url.scheme=="gce+ssh":
+            self.js = GCEService(lrms_saga_url, pilot_compute_description)
+        elif lrms_saga_url.scheme=="ec2+ssh" or lrms_saga_url.scheme=="euca+ssh" \
+            or lrms_saga_url.scheme=="nova+ssh":
+            self.js = EC2Service(lrms_saga_url, pilot_compute_description)
+        else:
+            self.js = SAGAJobService(lrms_saga_url)
         
-        self.number_nodes=int(number_nodes)        
-        
+        ##############################################################################
         # create job description
         jd = SAGAJobDescription()
         
@@ -228,17 +251,32 @@ class bigjob(api.base.bigjob):
             # if no working dir is set assume use home directory
             # will fail if home directory is not the same on remote machine
             # but this is just a guess to avoid failing
-            self.working_directory = os.path.expanduser("~") 
-            
-            
+            #self.working_directory = os.path.expanduser("~")
+            self.working_directory = "" 
+        
+        if queue != None:
+            jd.queue = queue
+        if project !=None:
+            jd.project=project       
+        if walltime!=None:
+            if is_bliss:
+                jd.wall_time_limit=int(walltime)
+            else:
+                jd.wall_time_limit=str(walltime)
+    
+        
+        ##############################################################################
+        # File Management and Stage-In            
         # Determine whether target machine use gsissh or ssh to logon.
         # logger.debug("Detect launch method for: " + lrms_saga_url.host)        
         # self.launch_method = self.__get_launch_method(lrms_saga_url.host,lrms_saga_url.username)
-        
-        ##############################################################################
-        # File Management and Stage-In
         self.bigjob_working_directory_url=""
-        if lrms_saga_url.scheme.startswith("condor")==False:           
+        if lrms_saga_url.scheme.startswith("gce") or lrms_saga_url.scheme.startswith("ec2")\
+            or lrms_saga_url.scheme.startswith("euca") or lrms_saga_url.scheme.startswith("nova"):
+            logger.debug("File Staging for Cloud Instances currently not supported.")
+        elif lrms_saga_url.scheme.startswith("condor") == True:
+            logger.debug("Using Condor file staging")
+        else:           
             # build target url for working directory
             # this will also create the remote directory for the BJ
             # Fallback if working directory is not a valid URL
@@ -272,6 +310,12 @@ class bigjob(api.base.bigjob):
             
             logger.debug("BJ Working Directory: %s", self.working_directory)      
         
+        if lrms_saga_url.scheme.startswith("condor")==False:
+            jd.working_directory = self.working_directory
+        else:
+            jd.working_directory=""
+        
+        
         
         ##############################################################################
         # Create and process BJ bootstrap script
@@ -292,6 +336,8 @@ class bigjob(api.base.bigjob):
             elif lrms_saga_url.scheme == "ssh":
                 bootstrap_script = self.__escape_ssh(bootstrap_script)
         logger.debug(bootstrap_script)
+        
+        
         # Define Agent Executable in Job description
         # in Condor case bootstrap script is staged 
         # (Python app cannot be passed inline in Condor job description)
@@ -327,45 +373,16 @@ class bigjob(api.base.bigjob):
             jd.spmd_variation = "single"
             jd.arguments = ["python", "-c", bootstrap_script]
             jd.executable = "/usr/bin/env"           
-       
-        if queue != None:
-            jd.queue = queue
-        if project !=None:
-            jd.project=project       
-        if walltime!=None:
-            if is_bliss:
-                jd.wall_time_limit=int(walltime)
-            else:
-                jd.wall_time_limit=str(walltime)
-    
-        if lrms_saga_url.scheme.startswith("condor")==False:
-            jd.working_directory = self.working_directory
-        else:
-            jd.working_directory=""
-
+      
         logger.debug("Working directory: " + jd.working_directory)
         
         jd.output = os.path.join(self.working_directory, "stdout-" + self.uuid + "-agent.txt")
         jd.error = os.path.join(self.working_directory, "stderr-" + self.uuid + "-agent.txt")
-           
-        # Submit job
-        js = None    
-        if userproxy != None and userproxy != '':
-            s = SAGASession()
-            os.environ["X509_USER_PROXY"]=userproxy
-            ctx = SAGAContext("x509")
-            ctx.set_attribute ("UserProxy", userproxy)
-            s.add_context(ctx)
-            logger.debug("use proxy: " + userproxy)
-            js = SAGAJobService(s, lrms_saga_url)
-        else:
-            logger.debug("use standard proxy")
-            js = SAGAJobService(lrms_saga_url)
-
+      
+        ##############################################################################
+        # Create and submit pilot job to job service
         logger.debug("Creating pilot job with description: %s" % str(jd))
-              
-
-        self.job = js.create_job(jd)
+        self.job = self.js.create_job(jd)
         logger.debug("Submit pilot job to: " + str(lrms_saga_url))
         self.job.run()
         return self.pilot_url
@@ -515,6 +532,7 @@ class bigjob(api.base.bigjob):
                             #logger.debug("Add attribute: " + str(i) + " Value: " + jd.get_attribute(i))
                             job_dict[i] = jd.get_attribute(i)
                 
+                # Other pilot state information
                 job_dict["state"] = str(Unknown)
                 job_dict["job-id"] = str(job_id)
                 logger.debug("job dict: " + str(job_dict))
@@ -1041,7 +1059,45 @@ def environment():
         self._environment = None
     return locals()
 
+
+def input_data():
+    doc = "List of input data units."
+    def fget(self):
+        return self._input_data
+    def fset(self, val):
+        self._input_data = val
+    def fdel(self, val):
+        self._input_data = None
+    return locals()
+
+def output_data():
+    doc = "List of output data units."
+    def fget(self):
+        return self._output_data
+    def fset(self, val):
+        self._output_data = val
+    def fdel(self, val):
+        self._output_data = None
+    return locals()
+
      
 class description(bliss.saga.job.Description):
     """ Sub-job description """
-    environment = property(**environment())   
\ No newline at end of file
+    environment = property(**environment())   
+    input_data = property(**input_data())
+    output_data = property(**output_data())   
+    
+    
+    def __init__(self):
+        bliss.saga.job.Description.__init__(self)
+        #self.attributes_extensible_ (True)
+        
+        # Extend description class by Pilot-Data relevant attributes
+        self._output_data = None
+        self._input_data = None
+        
+        self._register_rw_vec_attribute(name="InputData", 
+                                        accessor=self.__class__.input_data) 
+        self._register_rw_vec_attribute(name="OutputData", 
+                                        accessor=self.__class__.output_data) 
+        
\ No newline at end of file
